#ifndef HASHTABLE_HH
#define HASHTABLE_HH

#include <vector>
#include <utility> // std::pair
#include <iostream> // std::cout used in print

/*
 * A simple hash table implementation with chaining.
 *   Parameters are the number of buckets and a hash function.
 *   Hash function can be anything that has () operator.
 *   For an element e, the hash table accesses bucket _func(e) % _table.size()
 *   where _func is the hash function and _table.size() the number of buckets in
 *   the hash table.
 * The table itself is implemented as a vector of of linked-lists where
 *   the linked-lists are Node objects that are connected via their _next pointer.
 * Summary member functions:
 * insert() allows to insert an element in the hash table.
 * getBucket() returns a pointer to a specific bucket.
 * getBucketIter returns an iterator to the bucket specified by the parameter idx.
 *   See BucketIterator class for details.
 * getTable() returns a reference to the table.
 * getAverageAndLongestBucketLength() returns the average and
 *   longest bucket length in the table.
 * noBuckets() returns the number of buckets in the table.
 */

template <typename Tdata, typename Thashfunc>
class HashTable {
  public:
    struct Node {
      Node (Tdata aVal, Node* aNext)
        : _val(aVal), _next(aNext) {
        }
      Node (const Node&) = delete; // surpresses -Weffc++ warning
      Node& operator= (const Node&) = delete; // surpresses -Weffc++ warning
      // member variables
      Tdata _val;
      Node* _next;
    };
  public:
    class BucketIterator {
      public:
        BucketIterator (const Node* aNode) 
        : _node(aNode)
        {}
      public:
        BucketIterator& operator++ () {
          _node = _node->_next;
          return *this;
        }
        const Tdata& operator* () const {
          return _node->_val;
        }
        bool isValid () const {
          return _node != NULL;
        }
      private:
        const Node* _node;
    };
  public:
    // constructor
    HashTable (unsigned aNoBuckets, const Thashfunc& aFunc)
      : _func(aFunc), _table(aNoBuckets, NULL) 
    {}
  public:
    // destructor
    ~HashTable () { // destructor
      for (Node* node : _table) {
        while (node) {
          Node* currnode = node;
          node = node->_next;
          delete currnode;
        }
      }
    }
  public:
    void insert (const Tdata& value) {
      unsigned hashval = _func(value);
      hashval %= _table.size(); // safe guard
      Node* newnode = new Node (value, _table[hashval]);
      _table[hashval] = newnode;
    }
  public:
    const Node* getBucket (const Tdata& value) const {
      unsigned hashval = _func(value);
      hashval %= _table.size(); // safe guard
      return _table[hashval];
    }
    BucketIterator getBucketIter (unsigned idx) const {
      return BucketIterator(_table.at(idx));
    }
    const std::vector<Node*>& getTable () const {
      return _table;
    }
  public:
    std::pair<double, unsigned> getAverageAndLongestBucketLength () const {
      unsigned noUsedBuckets = 0;
      unsigned sumLengthBuckets = 0;
      unsigned lengthLongestBucket = 0;
      for (Node* node : _table) {
        if (node) {
          ++noUsedBuckets;
          unsigned lengthThisBucket = 0;
          while (node) {
            node = node->_next;
            ++lengthThisBucket;
          }
          sumLengthBuckets += lengthThisBucket;
          if (lengthLongestBucket < lengthThisBucket) {
            lengthLongestBucket = lengthThisBucket;
          }
        }
      }
      double averageLengthBucket = (double) sumLengthBuckets / noUsedBuckets;
      return {averageLengthBucket, lengthLongestBucket};
    }

    unsigned noBuckets () const {
      return _table.size();
    }

    /*
    void print () const {
      for (unsigned i = 0; i < _table.size(); ++i) {
        std::cout << "bucket[" << i << "]: ";
        Node* node = _table[i];
          while (node) {
            std::cout << node->_val << " -> ";
            node = node->_next;
          }
        std::cout << std::endl;
      }
    }
    */
  private:
    Thashfunc          _func;
    std::vector<Node*> _table;
};

#endif
