#include <random>
#include <cassert> // assert
#include <string>
#include <cstdint> // defines a set of integral type with specific width requirements
#include <fstream> // to read files
#include <iostream>
#include <iomanip>

#include "hashtable.hh"

/*
 * This files implements different hash functions.
 * For 32 bit integer:
 *   - DivisionHash implements (a*x + b) mod p
 *   - PolynomialHash implements (a1*x1 + a2*x2 + a3*x3 + a4*x4) % p,
 *     where xi refers i-th byte of some input integer x and ai are coefficients
 *   - murmurHash implements murmur hashing as in script
 * For strings:
 *   - LarsonStringHash implements the string hash function by Paul Larson
 *   - sumStringHash implements a hash function that simply returns the sum of numeric values
 *     of all characters in a string.
 * In the main() function we check on the average bucket length and maximum bucket length
 * given a hash function and given some data. Note that the average bucket length is
 * computed over the non-empy buckets only.
 * For integers, the data is uniformly distributed numbers.
 * For strings, the data is the 1000 most common words in the English language, cf. words.txt.
 */


struct DivisionHash {
  unsigned operator() (std::int32_t elem) {
    return (a*elem + b) % p;
  }
  int a;
  int b;
  int p;
};

struct PolynomialHash {
  unsigned operator() (std::int32_t elem) {
    int8_t* elemBytePointer= (int8_t*) &elem;
    int8_t x1, x2, x3, x4;
    x1 = *elemBytePointer;
    x2 = *(elemBytePointer+1);
    x3 = *(elemBytePointer+2);
    x4 = *(elemBytePointer+3);
    return (a1*x1 + a2*x2 + a3*x3 + a4*x4) % p;
  }
  int a1, a2, a3, a4;
  int p;
};

unsigned murmurHash (std::int32_t elem) {
  int x = elem;
  x ^= x >> 16;
  x *= 0x85ebca6b;
  x ^= x >> 13;
  x *= 0xc2b2ae35;
  x ^= x >> 16;
  return x;
}


struct LarsonStringHash {
  unsigned operator() (std::string word) {
    unsigned h = salt;
    const char* s = word.data();
    while (*s) {
      h = h * 101 + *s++;
    }
    return h;
  }
  unsigned salt;
};


unsigned sumStringHash (std::string word) {
  unsigned sum = 0;
  for (unsigned i = 0; i < word.size(); ++i) {
    sum += word[i];
  }
  return sum;
}


// function to test the average bucket length and maximum bucket length
// of a given hash table using random integer data
template <typename Thashtable>
void runInt (Thashtable& ht, unsigned noInserts) {
  std::default_random_engine generator;
  std::uniform_int_distribution<std::int32_t> distribution;

  for (unsigned i = 0; i < noInserts; ++i) {
    std::int32_t randval = distribution(generator);
    ht.insert(randval);
  }

  std::pair<double, unsigned> pair = ht.getAverageAndLongestBucketLength();
  std::cout << "average bucket length: "      << std::setw(7) << pair.first
            << "  |  longest bucket length: " << std::setw(4) << pair.second
            << "  |  number of buckets in table: " << ht.noBuckets()
            << std::endl;
}


// function to test the average bucket length and maximum bucket length
// of a given hash table using string data as given in words.txt
template <typename Thashtable>
void runString (Thashtable& ht) {
  // read file
  std::vector<std::string> words;
  std::ifstream file ("words.txt"); // contains the 1000 most common words in English
  std::string word;
  while (std::getline(file, word)) {
    words.push_back(word);
  }

  // insert words into hash table
  for (std::string& word : words) {
    ht.insert(word);
  }

  std::pair<double, unsigned> pair = ht.getAverageAndLongestBucketLength();
  std::cout << "average bucket length: "      << std::setw(7) << pair.first
            << "  |  longest bucket length: " << std::setw(4) << pair.second
            << "  |  number of buckets in table: " << ht.noBuckets()
            << std::endl;
}


int main (int argc, char* argv[]) {
  if (argc != 2) {
    std::cout << "usage: " << argv[0]
              << " <division | polynomial | murmur | larsonString | sumString>"
              << std::endl;
    return 111;
  }

  // initialize hash function
  DivisionHash division;
  division.a = 1;
  division.b = 0;
  division.p = 1000;

  // initialize hash function
  PolynomialHash polynomial;
  polynomial.a1 = 3;
  polynomial.a2 = 4;
  polynomial.a3 = 5;
  polynomial.a4 = 6;
  polynomial.p = 1000;

  // initialize hash function
  LarsonStringHash larson;
  larson.salt = 0;

  unsigned noInserts = 10'000; // number of inserts. Only used for runInt()

  std::string hashtype = argv[1];
  std::cout << "Hash type is " << std::setw(12) << hashtype << ":  ";
  if (hashtype == "division") {
    unsigned noBuckets = noInserts;
    HashTable<std::int32_t, DivisionHash> htDiv (noBuckets, division);
    runInt(htDiv, noInserts);
  } else if (hashtype == "polynomial") {
    unsigned noBuckets = noInserts;
    HashTable<std::int32_t, PolynomialHash> htPoly (noBuckets, polynomial);
    runInt(htPoly, noInserts);
  } else if (hashtype == "murmur") {
    unsigned noBuckets = noInserts;
    using murmur_f = unsigned (*) (int); // type of function pointer
    HashTable<std::int32_t, murmur_f> htMurmur (noBuckets, murmurHash);
    runInt(htMurmur, noInserts);
  } else if (hashtype == "larsonString") {
    unsigned noBuckets = 1000;
    HashTable<std::string, LarsonStringHash> htLarson (noBuckets, larson);
    runString(htLarson);
  } else if (hashtype == "sumString") {
    unsigned noBuckets = 1000;
    using sumString_f = unsigned (*) (std::string);
    HashTable<std::string, sumString_f> htSumString (noBuckets, sumStringHash);
    runString (htSumString);
  } else {
    std::cout << "unknown hash type" << std::endl;
  }
}
